#include "Instance.h"

/******************************************************************************
  Method:       new 
  Class:        Instance
  Arguments:    SequenceObject * pSeqObj, strand_modes s, int start, int length
  
  Description:  constructor
							  create new instance from the given sequence
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
Instance::Instance(SequenceObject * pSeqObj, strand_modes s, int start,
                   int length)
{
  // FC
  _pMotifID = NULL;
  _pOverlaps = NULL;
  _pJoinMotifIDs = NULL;
	
  _pId = NULL;
  // set parameters
  if (start < 0 || length < 0)
  {
    _start = 0;
    _length = 0;
    _score = 0;
    _pSite = NULL;
    _pSiteString = NULL;
    _pSeqName = NULL;
    return;
  }

  // set internal parameters
  _strand = s;
  _start = start;
  _length = length;
  _pParentSequence = pSeqObj;
  _pSeqName = new string(*(pSeqObj->GetID()));

  // initial score is 0
  _score = 0;

  // get site
  _pSite = _pParentSequence->GetSubSequenceVector(s, start, length);

  // translate sequence vector to 
  if (_pSite == NULL)
  {
    _pSiteString = NULL;
    cerr <<
      "-Error-- Instance(): Unable to get site from sequence => No instance created "
      << endl;
  }
  else
  {

    _pSiteString = new string;

    for (uint i = 0; i < _pSite->size(); i++)
    {
      switch ((*_pSite)[i])
      {
      case 0:
        _pSiteString->append("A");
        break;
      case 1:
        _pSiteString->append("C");
        break;
      case 2:
        _pSiteString->append("G");
        break;
      case 3:
        _pSiteString->append("T");
        break;
      default:
        _pSiteString->append("N");
      }
    }
    // cerr << "Instance " << index << " => created string: " << *_pSiteString << endl;
  }
}
/******************************************************************************
  Method:       new
  Class:        Instance
  Arguments:    Instance& copyInstance
  
  Description:  copy constructor
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
Instance::Instance(Instance& copyInstance)
{
  // FC
  _pMotifID = NULL;
  _pOverlaps = NULL;
  _pJoinMotifIDs = NULL;
  // _pId is not copied 
  _pId = NULL;

  // set internal parameters
  _strand = copyInstance.Strand();
  _start = copyInstance.Start();
  _length = copyInstance.Length();
  _score = copyInstance.Score();

  _pSiteString = new string(*(copyInstance.PrintSite()));

  if (copyInstance.ParentSequence() != NULL) //
  {
    _pParentSequence = copyInstance.ParentSequence();
    _pSeqName = new string(*(_pParentSequence->GetID()));
    _pSite = _pParentSequence->GetSubSequenceVector(_strand, _start, _length);
  }
  else //
  {
    _pParentSequence = NULL; 
    _pSeqName = new string(*(copyInstance.PrintSeqName())); 
    _pSite = NULL; 
    Sequence * copy = copyInstance.Site(); 
    if ( copy != NULL) //
    {
      _pSite = new Sequence;   
      for(int i = 0; i < (int)copy->size(); i++)  
        _pSite->push_back((*copy)[i]); 
    }
  }
}
/******************************************************************************
  Description:  constructor - for FuzzyClustering
  Author_Date:  MC_2009/07/22
******************************************************************************/
Instance::Instance(string * seqid, const int &start, const int &end, 
           const strand_modes &s, string * site, string motifid, double w)
{ 
  // cerr << "debug - Instance::Instance - begin" << endl;

  _length = end - start + 1;// 
  // set parameters
  if (start < 0 || _length < 0)
  {
    _start = 0;
    _length = 0;
    _score = 0;
    _pSite = NULL;
    _pSiteString = NULL;
    _pSeqName = NULL;
    //cerr << "--Error--Instance::Instance - invalid input" << endl;
  }
  // set internal parameters
  _strand = s;
  _start = start;
  _pParentSequence = NULL; //
  _pId = NULL;
  _pSeqName = new string(*seqid);//

  _weight = w;

  _pMotifID = new string(motifid);//
  _pOverlaps = NULL;
  _pOcc = 1;
  _pJoinMotifIDs = new vector<string>;
  _pJoinMotifIDs->push_back(*_pMotifID);
  _pJoinWeights = new vector<double>;
  _pJoinWeights->push_back(_weight);

  // initial score is 0
  _score = 0;
  // score used in CompareID()class
  _iStore = 0;

  // get site
  _pSiteString = new string(*site);//

  // translate to sequence (0123) code
  _pSite = NULL; // this must first be set NULL !
  _TranslateToSequence(); // _pSite
  //cerr << "debug - Instance::Instance - end" << endl;
}
/******************************************************************************
  Method: 
  Class:        Instance
  Arguments:    none
  
  Description:  destructor
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
Instance::~Instance()
{
  // FC
  if (_pOverlaps != NULL)
  {
    for(int i = 0; i < (int)_pOverlaps->size(); i++)
    {  // delete instance
      if ((*_pOverlaps)[i] != NULL) delete (*_pOverlaps)[i];
      (*_pOverlaps)[i] = NULL;
    }
    delete _pOverlaps; // delete vector
  }
  _pOverlaps = NULL; 

  if (_pJoinMotifIDs != NULL)
    delete _pJoinMotifIDs;
  _pJoinMotifIDs = NULL;

  if (_pJoinWeights != NULL)
    delete _pJoinWeights;
  _pJoinWeights = NULL;

  if (_pMotifID != NULL) delete _pMotifID;_pMotifID = NULL;

  // reset pointer to string (should be defined outside this scope)
  // _pId = NULL;

  if (_pSite != NULL)
    delete _pSite;
  _pSite = NULL;

  if (_pSiteString != NULL)
    delete _pSiteString;
  _pSiteString = NULL;

  if ( _pSeqName != NULL )
    delete _pSeqName;
  _pSeqName = NULL;
}

/******************************************************************************
  Method:       CheckLeftExtension
  Class:        Instance
  Arguments:    int p
  
  Description:  Check if the instance is extensible p positions to the left
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool Instance::CheckLeftExtension(int pos){
  if ( _start - pos < 0 ){
    cerr << "--Warning-- Instance::ExtendLeft(): Unable to extend site to left." << endl;
    return false;
  }
  
  return true;
}


/******************************************************************************
  Method:       ExtendLeft
  Class:        Instance
  Arguments:    int p
  
  Description:  Extend the instance p positions to the left
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
Instance* Instance::ExtendLeft(int pos){
  // change start position ans site length
  _start = _start - pos;
  _length += pos;
  
  // update site
  if (_pSite != NULL)
    delete _pSite;
  // cerr << "DEBUG: Instance::ExtendLeft(int pos) " << pos << " -> " << _start << " " << _length << endl;
  _pSite = _pParentSequence->GetSubSequenceVector(_strand, _start, _length);
  
  // update site string
  _pSiteString->resize(_length);
  for (int i = 0; i < _length; i++)
    {
      switch ((*_pSite)[i])
      {
      case 0:
        _pSiteString->replace(i,1,"A");
        break;
      case 1:
        _pSiteString->replace(i,1,"C");
        break;
      case 2:
        _pSiteString->replace(i,1,"G");
        break;
      case 3:
        _pSiteString->replace(i,1,"T");
        break;
      default:
        _pSiteString->replace(i,1,"N");
      }
    }
  
  return this;
}


/******************************************************************************
  Method:       CheckRightExtension
  Class:        Instance
  Arguments:    int p
  
  Description:  Check if the instance is extensible p positions to the right
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
bool Instance::CheckRightExtension(int pos){
  if ( _start + _length + pos > _pParentSequence->Length() ){
    cerr << "--Warning-- Instance::ExtendRight(): Unable to extend site to right." << endl;
    return false;
  }
  
  return true;
}

  
/******************************************************************************
  Method:       ExtendRight
  Class:        Instance
  Arguments:    int p
  
  Description:  Extend the instance p positions to the right
  
  Date:         2003/06/26
  Author:       Gert Thijs <gert.thijs@esat.kuleuven.ac.be>
  
******************************************************************************/
Instance*  Instance::ExtendRight(int pos){
  // change length of the site
  _length += pos;
  
  // update site
  if (_pSite != NULL)
    delete _pSite;

  // cerr << "DEBUG: Instance::ExtendRight(int pos) " << pos << " -> " << _start << " " << _length << endl;
  _pSite = _pParentSequence->GetSubSequenceVector(_strand, _start, _length);

  for (int i = _length-pos; i < _length; i++)
    {
      switch ((*_pSite)[i])
      {
      case 0:
        _pSiteString->append("A");
        break;
      case 1:
        _pSiteString->append("C");
        break;
      case 2:
        _pSiteString->append("G");
        break;
      case 3:
        _pSiteString->append("T");
        break;
      default:
        _pSiteString->append("N");
      }
    }
  return this;
}

/******************************************************************************
  Description:  InstanceID() : compose with actual information, new string
  Author_Date:  MC_2009/07/22
******************************************************************************/
string *
Instance::InstanceID()
{
  ostringstream instanceid;
  instanceid << *_pSeqName << "_" << _start << "_" << _start + _length -1 << "_"; 
  if (_strand == plus_strand) instanceid << "+"; else instanceid << "-" ;
  if (_pSiteString != NULL) 
  {instanceid << "_"  << *_pSiteString;}
  else {instanceid << "_NULL";}
  string * pInstID = new string(instanceid.str());
  instanceid.flush();
  return pInstID;
}

/******************************************************************************
  Description:  SimilarMotifs : count number of similars in _pJoinMotifIDs
                from two instances
  // function is used in BuildPairwiseMatrix
  Author_Date:  MC_2009/07/22
******************************************************************************/
int 
Instance::SimilarMotifs(Instance * pInstance2)
{
  int count = 0;
  // get ids from instance 2
  vector<string> * pMotifs2 = pInstance2->GetAllMotifs();

  int N = (int)_pJoinMotifIDs->size();
  int M = (int)pMotifs2->size();
  // compare both (same ascending store order (push_back was used)
  // ...computationally not efficient but acceptable for expected sizes of _C1
  string id1, id2;
  for(int i = 0; i < N; i++)
  {
    id1 = (*_pJoinMotifIDs)[i]; 
    for(int j = 0; j < M; j++)
    {
      id2 = (*pMotifs2)[j]; 
      if ( !(id1).compare(id2)) 
      { count++; break;}
    }
  }

  return count;
}

/******************************************************************************
  Description:  GetOverlaps() : return instancs-ids of all instances in 
                _pOverlaps // new string
  Author_Date:  MC_2009/12/10
******************************************************************************/
string *
Instance::GetOverlaps()
{
  // the instances in _pOverlaps are assumed single
  if (_pOverlaps == NULL) return NULL;
  string * pIDs(NULL), *pid(NULL);
  ostringstream ids;
  int occ = 1;
  double weight = (*_pOverlaps)[0]->GetWeight();
  string * previousid = (*_pOverlaps)[0]->InstanceID();
  bool komma = false;
  for(int i = 1; i <= (int)_pOverlaps->size(); i++)
  {
    // "previous" instance can be printed now
    if (pid != NULL)
    {
      if (komma){ ids << ",";}
      // add to reporting string
      ids << *previousid << "(" << occ << "," << setprecision(2) << weight/occ << ")";
      komma = true;  
      weight = (*_pOverlaps)[i-1]->GetWeight(); // reset
      occ = 1;
      delete previousid; previousid = pid; pid = NULL;
    }
    if (i == (int)_pOverlaps->size())
    {
      // process last one
      if (komma){ ids << ",";}
      ids << *previousid << "(" << occ << "," << setprecision(2) << weight/occ << ")";
      delete previousid; previousid = NULL;
      if (pid != NULL) delete pid; pid = NULL;
      break;
    }
    // read the current instance
    pid = (*_pOverlaps)[i]->InstanceID();
    // evaluate if exact with previous 
    if (*pid == *previousid)
    {
      occ++;
      weight += (*_pOverlaps)[i]->GetWeight();
      delete pid; pid = NULL; // keep previousid
    }
  }
  // print
  pIDs = new string(ids.str());
  ids.flush();
  return pIDs;
}

/******************************************************************************
  Description:  TranslateToSequence() (acgt->0123) 
  Author_Date:  MC_2009/07/27
******************************************************************************/
void 
Instance::_TranslateToSequence()
{
  if (_pSite != NULL); delete _pSite;
  _pSite = new Sequence;
  int i = 0;
  while ( i < _length)
  {
    switch (_pSiteString->at(i))
    {
    case 'a':
    case 'A':
      _pSite->push_back(0);
      break;
    case 'c':
    case 'C':
      _pSite->push_back(1);
      break;
    case 'g':
    case 'G':
      _pSite->push_back(2);
      break;
    case 't':
    case 'T':
      _pSite->push_back(3);
      break;
    default:
      _pSite->push_back(-1);
    }
    i++;
  }
  return;
}
/******************************************************************************
  Description:  MinStart() : return the instance from _Overlaps with earliest 
                _start
  Author_Date:  MC_2009/07/26
******************************************************************************/
Instance *
Instance::MinStart()
{
  if (_pOverlaps == NULL) return NULL;
  
  int minimum = _start;
  Instance * minInstance = NULL;

  for(int i = 0; i < (int)_pOverlaps->size(); i++)
  {
    if ((*_pOverlaps)[i]->Start() < minimum)
    { 
      minInstance = (*_pOverlaps)[i];
      minimum = (*_pOverlaps)[i]->Start();
    }
  }
  return minInstance; 
}

/******************************************************************************
  Description:  MaxEnd() : return the instance from _Overlaps with latest end
  Author_Date:  MC_2009/07/26
******************************************************************************/
Instance *
Instance::MaxEnd()
{
  if (_pOverlaps == NULL) return NULL;

  int maximum = _start + _length;
  Instance * maxInstance = NULL;

  for(int i = 0; i < (int)_pOverlaps->size(); i++)
  {
    if ((*_pOverlaps)[i]->Start() + (*_pOverlaps)[i]->Length() > maximum)
    { 
      maxInstance = (*_pOverlaps)[i];
      maximum = (*_pOverlaps)[i]->Start() + (*_pOverlaps)[i]->Length();
    }
  }
  return maxInstance; 
}

/******************************************************************************
  Description:  CreateOverlaps (_pOverlaps) -> artificial instance
  Author_Date:  MC_2011/09/21
******************************************************************************/
void 
Instance::CreateOverlaps()
{
  _pOverlaps = new vector<Instance*>;
  Instance * pItself = new Instance(_pSeqName, _start, _start + _length -1,
                                   _strand, _pSiteString, *_pMotifID, _weight);
  _pOverlaps->push_back(pItself);
  // the other variables reamin the same 
  // so the artifical still describes the single instance at this stage
  return;
}
/******************************************************************************
  Description:  AddInstance : 
                add instance to _pOverlaps and _pJoinMotifIDs, and update 
                other variables
   // this function replaces earlier code (UpdateWithOverlap, 
      UpdateWithExact, ConvertToExact)
  Author_Date:  MC_2011/09/21
******************************************************************************/
void 
Instance::AddInstance(Instance * pInst)
{
  //cerr << "debug--Instance::AddInstance - begin" << endl;

  // pInst is single or exacts and Instance is already artifical
  _weight = (_weight*_pOcc);
  if (pInst->NbrOcc() == 1) 
  { _pOverlaps->push_back(pInst);
    // sum weight
    _weight += pInst->GetWeight();
  }
  else 
  {
    vector<Instance*> *pExacts = pInst->GetOverlapInsts();
    for (int i = 0; i < (int)pExacts->size(); i++)
    { _pOverlaps->push_back((*pExacts)[i]);
      _weight += (*pExacts)[i]->GetWeight();
    }
  }
  _pOcc += pInst->NbrOcc();
  // make an average weight
  _weight /= _pOcc;
  // set start and end to min/max
  int end = _start + _length -1;
  if (pInst->Start()+pInst->Length()-1 > end) 
  { end =  pInst->Start()+pInst->Length()-1;}
  if (pInst->Start() < _start) {_start = pInst->Start();}
  _length = end - _start +1;

  // reset other variables of the artificial (not relevant for now)
  if (_pMotifID != NULL) delete _pMotifID; _pMotifID = NULL;
  if (_pJoinMotifIDs != NULL) delete _pJoinMotifIDs; _pJoinMotifIDs = NULL;
  if (_pJoinWeights != NULL) delete _pJoinWeights; _pJoinWeights = NULL;
  if (_pSiteString != NULL) delete _pSiteString; _pSiteString = NULL;
  if (_pSite != NULL) delete _pSite; _pSite = NULL;

  return;
}

/******************************************************************************
  Description:  UpdateInstance : 
                make it non-artifical by updating all its variables 
  Author_Date:  MC_2011/09/21
******************************************************************************/
void 
Instance::UpdateInstance()
{
  //cerr << "debug--Instance::UpdateInstance - begin" << endl;

  if (_pOverlaps == NULL) 
  { cerr << "debug--Instance::UpdateInstance, error1" << endl;
    return;
  }
  if (_pOcc != (int)_pOverlaps->size())
  { cerr << "debug--Instance::UpdateInstance, error2" << endl;
    return;
  }

  // store motifids and weight in corresponding vector
  if (_pJoinMotifIDs != NULL) delete _pJoinMotifIDs; 
  _pJoinMotifIDs = new vector<string>;
  if (_pJoinWeights != NULL) delete _pJoinWeights; 
  _pJoinWeights = new vector<double>;

  // sequentially process info of each individual instance
  // this may also be only one single instance (_pOcc == 1)
  double sumW = 0; Instance * pSingle = NULL;
  double dstart(0), dend(0); 
  for(int i = 0; i < _pOcc; i++)
  {
    pSingle = (*_pOverlaps)[i];
    _pJoinMotifIDs->push_back(*pSingle->GetMotifID());
    _pJoinWeights->push_back(pSingle->GetWeight());
    sumW += (double)pSingle->GetWeight();
    dstart += (double)pSingle->GetWeight() * pSingle->Start();
    dend += (double)pSingle->GetWeight() * (pSingle->Start() + pSingle->Length() -1); 
  }
  // compute a weighted average start/length
  dstart /= sumW; 
  if (dstart - int(dstart) >= 0.5)
  {_start = int(dstart)+1;} else {_start = (int)dstart;}
  dend /= sumW; int end = 0;
  if (dend - int(dend) > 0.5)
  { end = int(dend)+1;} else {end = (int)dend;}
  _length = end - _start + 1;
  _weight= sumW/_pOcc; // averaged to use for RemoveLowProbIndividuals

  // and finally make the corresponding string
  if (_pSiteString != NULL) delete _pSiteString;
  if (_pSite != NULL) delete _pSite; _pSite = NULL;
  string * pJoined = new string(*(*_pOverlaps)[0]->PrintSite());
  int start1 = (*_pOverlaps)[0]->Start();
  int end1 = (*_pOverlaps)[0]->Start()+(*_pOverlaps)[0]->Length() -1;
  string::size_type posx = string::npos; int shift;
  for(int i = 1; i <= _pOcc; i++) // will be skipped in case of _pOcc = 1
  {
    posx = pJoined->find("x");
    if (start1 <= _start && end1 >= end
            && posx == string::npos)
    { // finalize the string
      _pSiteString = new string(pJoined->substr(_start-start1,_length));
      break;
    }
    if (i == _pOcc)
    { cerr << "debug--Instance::UpdateInstance, error3" << endl;}
    pSingle = (*_pOverlaps)[i]; // is a single instance
    // adjust xx positions if applicable/possible
    while ( (posx = pJoined->find("x")) != string::npos)
    {
      shift = pSingle->Start() - start1;
      int p = 0;
      for (; p < (int)pSingle->Length(); p++)
      {
        if ( p + shift == (int)posx)
        {
          (*pJoined)[posx] = (*pSingle->PrintSite())[p];
        }
      }
      if ( p == pSingle->Length()) { break;} 
    }
    // add preceeding or trailing info if applicable
    if (start1 > _start && pSingle->Start() < start1)
    { 
      // insert also dummy's if so
      if ((pSingle->Start()+pSingle->Length()) < start1)
      {
        pJoined->insert(0, 
				start1-(pSingle->Start()+pSingle->Length()), 'x');
        pJoined->insert(0, *pSingle->PrintSite());
      }
      else 
      { // insert non-overlapping preceeding part 
        pJoined->insert(0, *pSingle->PrintSite(),
						0, start1 - pSingle->Start());
      }
      start1 = pSingle->Start();
    }
    if (end1 < end && pSingle->Start()+pSingle->Length()-1 > end1)
    { 
      // add also dummy's if so
      if ( pSingle->Start() > end1)
      {
        pJoined->append(
				pSingle->Start()-end1-1, 'x');
        pJoined->append(*pSingle->PrintSite());
      }
      else 
      { // add non-overlapping trailing part 
        pJoined->append(*pSingle->PrintSite(),
						end1- pSingle->Start(),
						pSingle->Start()+pSingle->Length()-1 - end1);
      }
      end1 = pSingle->Start()+pSingle->Length()-1;
    }
  }
  if (pJoined != NULL) delete pJoined; pJoined = NULL;
  // check
  if ((int)_pSiteString->size() != _length)
  { 
    cerr << "debug--Instance::UpdateInstance, error in string" << endl;
  }
  // and update _pSite
  else
  { 
    _TranslateToSequence();
  }

  // leave _pOverlaps as it is and _pMotifID stays unused
  if (_pMotifID != NULL) delete _pMotifID; _pMotifID = NULL;

  // except if the artifical was an individual instance after all
  if ( _pOcc == 1) 
  {
    _pMotifID = new string((*_pJoinMotifIDs)[0]);
    // delete the second layer
    delete (*_pOverlaps)[0]; // delete instance
    delete _pOverlaps; _pOverlaps = NULL; // delete vector
  }
  //cerr << "debug--Instance::UpdateInstance - end" << endl; 
  return;
}

